#ifndef ROBUST_MEASURE_HGFDBGRYTNGFSVHDHNDFGBJDVHFDH
#define ROBUST_MEASURE_HGFDBGRYTNGFSVHDHNDFGBJDVHFDH

#include <boost/container/flat_set.hpp>
#include <cmath>
#include <limits>
#include <algorithm>
#include <set>

template<typename func_t>
double robust_measure(func_t func, double tolerance = .05, int min_runs = 25, int max_runs = std::numeric_limits<int>::max()){
	boost::container::flat_multiset<double> measurements;
	std::multiset<double> buffer;
	using iter_t = typename decltype(measurements)::iterator;

	double glob_sum_x = 0.0;

	auto rel_stddev = [&](iter_t first, iter_t last, double sum_x){
		double actual_mean = sum_x / (last - first);
		double n = 0.0;
		double mean = 0.0;
		double m2 = 0.0;

		for(auto it = first; it != last; ++it){
			n++;
			double x = *it - actual_mean;
			double delta = x - mean;
			mean += delta/n;
			m2 += delta*(x - mean);
		}
		double variance = m2/n;
		double pop_stddev = std::sqrt(variance/n);

		double ret = pop_stddev / actual_mean;
		return ret;
	};

	double next_check = min_runs;

	while(measurements.size() + buffer.size() < max_runs){
		double result = func();
		buffer.insert(result);
		glob_sum_x += result;

		if(measurements.size() + buffer.size() == long(next_check)){
			boost::container::flat_multiset<double> new_measurements;
			auto it1 = measurements.begin();
			auto it2 = buffer.begin();
			while( it1 != measurements.end() && it2 != buffer.end() ){
				if(*it1 < *it2){
					new_measurements.insert(new_measurements.end(), *(it1++));
				}
				else{
					new_measurements.insert(new_measurements.end(), *(it2++));	
				}
			}
			if( it1 != measurements.end() ){
				new_measurements.insert(it1, measurements.end());
			}
			if( it2 != buffer.end() ){
				new_measurements.insert(it2, buffer.end());
			}
			measurements = std::move(new_measurements);
			buffer.clear();

			auto first = measurements.begin();
			auto last = measurements.end();
			double sum_x = glob_sum_x;

			while(last - first >= min_runs){
				if(rel_stddev(first, last, sum_x) < tolerance){
					return sum_x/(last - first);
				}
				sum_x -= *first;
				++first;
				--last;
				sum_x -= *last;
			}
			next_check *= 1.1;
		}
	}
	return glob_sum_x / measurements.size();
}
#endif
