#include "use_cases.hpp"
#include "sets/adaptive_slack_set"
#include "sets/comparable_hash_set"
#include "sets/comparable_hash_set2"
#include "sets/merge_set"
#include "sets/proxy_slack_set"
#include "sets/slack_set"
#include "sets/b_plus_set"
#include "sets/sparse_set"
#include "sets/circular_set"
#include <boost/container/flat_set.hpp>
#include <vector>
#include <string>

using std::string;
using std::set;

template<typename T> class custom_flat_set{}; 
template<> class custom_flat_set<int> : public sets::custom_flat_set<int>{};
template<> class custom_flat_set<std::string> : public sets::custom_flat_set<std::string>{};
template<> class custom_flat_set<record<50>> : public sets::custom_flat_set<record<50>>{};
template<typename T> struct needs_sorted_input<custom_flat_set<T>> : public std::true_type {};

using boost::container::flat_set;
using sets::circular_set;
template<typename T> struct needs_sorted_input<boost::container::flat_set<T>> : public std::true_type {};
template<typename T> struct needs_sorted_input<sets::circular_set<T>> : public std::true_type {};

template<typename T>
class hinted_flat_set{
	public:
		using iterator = typename boost::container::flat_set<T>::iterator;
	
		void clear(){ cont.clear(); }
		iterator begin(){ return cont.begin(); }
		iterator end(){ return cont.end(); }
		int size(){ return cont.size(); }
		template<typename U>
		iterator find(const U& u){ return cont.find(u); }
		template<typename U>
		int erase(const U& u){ return cont.erase(u); }

		template<typename U>
		std::pair<iterator, bool> insert(U&& u){
			iterator hint = sets::fast_lower_bound( cont.begin(), cont.end(), u );
			if(hint != end() && *hint == u){
				return {hint, false};
			}
			auto ret = cont.insert(hint, std::forward<U>(u));
			return {ret, true};
		}
	private:
		boost::container::flat_set<T> cont;
};
template<typename T> struct needs_sorted_input<hinted_flat_set<T>> : public std::true_type {};

template<typename T> class merge_set{}; 
template<> class merge_set<int> : public sets::merge_set<int, 1, 10, 200>{};
template<> class merge_set<std::string> : public sets::merge_set<std::string, 1, 10, 200>{};
template<> class merge_set<record<50>> : public sets::merge_set<record<50>, 1, 10, 100>{};

template<typename T> class slack_set{}; 
template<> class slack_set<int> : public sets::slack_set<int, 5, 10, 1000>{};
template<> class slack_set<std::string> : public sets::slack_set<std::string, 1, 2, 200>{};
template<> class slack_set<record<50>> : public sets::slack_set<record<50>, 1, 2, 250>{};

template<typename T> class proxy_slack_set{}; 
template<> class proxy_slack_set<int> : public sets::proxy_slack_set<int, 5, 10, 1000>{};
template<> class proxy_slack_set<std::string> : public sets::proxy_slack_set<std::string, 1, 2, 200>{};
template<> class proxy_slack_set<record<50>> : public sets::proxy_slack_set<record<50>, 1, 2, 250>{};

template<typename T> class adaptive_slack_set{}; 
template<> class adaptive_slack_set<int> : public sets::adaptive_slack_set<int, 9, 10, 9, 10, 2000>{};
template<> class adaptive_slack_set<std::string> : public sets::adaptive_slack_set<std::string, 9, 10, 9, 10, 600>{};
template<> class adaptive_slack_set<record<50>> : public sets::adaptive_slack_set<record<50>, 9, 10, 9, 10, 250>{};

template<typename T> class comparable_hash_set{}; 
template<> class comparable_hash_set<int> : public sets::comparable_hash_set<int, 8, 1000>{};
template<> class comparable_hash_set<std::string> : public sets::comparable_hash_set<std::string, 8, 200>{};
template<> class comparable_hash_set<record<50>> : public sets::comparable_hash_set<record<50>, 4, 100>{};

template<typename T> class comparable_hash_set2{}; 
template<> class comparable_hash_set2<int> : public sets::comparable_hash_set2<int, 8, 500>{};
template<> class comparable_hash_set2<std::string> : public sets::comparable_hash_set2<std::string, 8, 200>{};
template<> class comparable_hash_set2<record<50>> : public sets::comparable_hash_set2<record<50>, 8, 400>{};

template<typename T> class slide_set{}; 
template<> class slide_set<int> : public sets::slide_set<int, 9, 10>{};
template<> class slide_set<std::string> : public sets::slide_set<std::string, 9, 10>{};
template<> class slide_set<record<50>> : public sets::slide_set<record<50>, 8, 10>{};
template<typename T> struct needs_sorted_input<slide_set<T>> : public std::true_type {};

template<typename T> class b_plus_set{}; 
template<> class b_plus_set<int> : public sets::b_plus_set<int, 150, 500>{};
template<> class b_plus_set<std::string> : public sets::b_plus_set<std::string, 150, 150>{};
template<> class b_plus_set<record<50>> : public sets::b_plus_set<record<50>, 130, 150>{};

template<typename T> class sparse_set{};
template<> class sparse_set<int> : public sets::sparse_set<int>{};
template<> class sparse_set<std::string> : public sets::sparse_set<std::string>{};
template<> class sparse_set<record<50>> : public sets::sparse_set<record<50>>{};
